# The server part
# The place where i will forward the whole traffic
# By d4rk3r
import socket, argparse, threading, json, time
from app.utils.helpers import get_trace



class Server:
    def __init__(self, port):
        self.rooms = {}
        self.ip = '0.0.0.0'
        self.port = port

        while True:
            try:
                self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.s.bind((self.ip, self.port))

                self.s.listen()
                break
            except:
                print("[x] Couldn't bind to that port {}".format(str(self.port)))
                time.sleep(1)

        self.accept_connections()

    def accept_connections(self):

        print('[-] Running on IP: ' + self.ip)
        print('[-] Running on port: ' + str(self.port))

        while True:
            c, addr = self.s.accept()
            # Here we're going to handle clients with threads
            threading.Thread(target=self.handle_client,args=(c,addr,)).start()
            # A seperate thread to delete all old session tofree the ram
            threading.Thread(target=self.erase_old_sessions).start()

    def save_room_set(self, sock, json_data):
        """
        This method will create a session group and add the current id
        in it if everything is not set for that client

        """
        # if the session doesn't exist we create it
        if json_data["s"] not in self.rooms:
            self.rooms[json_data["s"]] = {"d": time.time()}

        # We check if a client_id already exist with that id
        # if self.rooms[json_data["s"]][json_data["i"]]

        # The limit of sessions to get a fluidity
        if len(self.rooms[json_data["s"]]) < 5:
            # if the current id not in the target session, we add it
            if json_data["i"] not in self.rooms[json_data["s"]]:
                self.rooms[json_data["s"]][json_data["i"]] = {"c": sock}

    def erase_old_sessions(self):
        """
        This method will loop over all current session and delete old one from 360
        just to free the RAM

        """

        for r in self.rooms:
            if time.time() - self.rooms[r]["d"] >= 60:
                print(f"[-] Session {r} erased !")
                del self.rooms[r]

        time.sleep(360)

    def broadcast(self, sock, json_data):
        """
        In this method we're going to broadcast the whole thing
        """
        self.save_room_set(sock, json_data)

        # We set a new time for the session cuz it still active
        self.rooms[json_data["s"]]["d"] = time.time()
        # We loop over all clients in the session
        # To send them the flux
        for elt in self.rooms[json_data["s"]]:
            try:
                client = self.rooms[json_data["s"]][elt]["c"]
                if client != self.s and client != sock:
                    try:
                        client.send(bytes(json.dumps(json_data), encoding="utf-8"))
                    except Exception as es:
                        get_trace()
            except Exception as es:
                pass

    def handle_client(self, c, addr):
        """
        To handle each socket client
        """
        while True:
            try:
                try:
                    json_data = json.loads(c.recv(2048).decode("utf-8"))

                    # i for the id, s for the session,
                    # 'v' for the video string and 'a' for the audio chunk
                    if all(k in json_data.keys() for k in ["i", "s"]):
                        self.broadcast(c, json_data)
                    else:
                        print("[x] Incomplete sequence !")
                except (json.decoder.JSONDecodeError, BrokenPipeError) as e:
                   pass
            except socket.error:
                c.close()


if __name__ == "__main__":
    # Initialize the arguments
    prs = argparse.ArgumentParser()
    prs.add_argument('-p', '--port',
            help='The port where the server will be running',
            type=int, default=1122)
    prs = prs.parse_args()

    # We start our server
    server = Server(prs.port)




from Crypto.Cipher import AES
from Crypto.Protocol.KDF import PBKDF2, HKDF
from Crypto.Hash import HMAC, SHA256, SHA512
from Crypto.Random import get_random_bytes
from Crypto.Util.Padding import pad, unpad

from base64 import b64encode, b64decode
from binascii import Error as Base64Error
from os.path import getsize as file_size
from re import sub as re_sub

#
#import time
#
#start = time.time()
#

class KriptaAES(object):
    '''
    Encrypts data and files using AES CBC/CFB - 128/192/256 bits.
    The encryption and authentication keys
    are derived from the supplied key/password using HKDF/PBKDF2.
    The key can be set either with `set_master_key` or with `random_key_gen`.
    Encrypted data format: salt[16] + iv[16] + ciphertext[n] + mac[32].
    Ciphertext authenticity is verified with HMAC SHA256.
    Requires pycryptodome https://pycryptodome.readthedocs.io
    :ivar key_iterations: int The number of PBKDF2 iterations.
    :ivar base64: bool Accepts ans returns base64 encoded data.
    :param mode: str Optional, the AES mode (cbc or cfb).
    :param size: int Optional, the key size (128, 192 or 256).
    :raises ValueError: if the mode or size is invalid.
    @author Sanix-darker
    '''
    def __init__(self, mode = 'CBC', size = 128):
        self._modes = {'CBC': AES.MODE_CBC, 'CFB': AES.MODE_CFB}
        self._sizes = (128, 192, 256)
        self._salt_len = 16
        self._iv_len = 16
        self._mac_len = 32
        self._mac_key_len = 32

        if mode.upper() not in self._modes:
            raise ValueError(mode + ' is not supported!')
        if size not in self._sizes:
            raise ValueError('Invalid key size!')
        self._mode = mode.upper()
        self._key_len = int(size / 8)
        self._master_key = None

        self.key_iterations = 20000
        self.base64 = True

    def encrypt(self, data, password = None):
        '''
        Encrypts data using the supplied password or a master key.
        The password is not required if a master key has been set -
        either with `random_key_gen` or with `set_master_key`.
        If a password is supplied, it will be used to create a key with PBKDF2.
        :param data: str or bytes or bytearray The plaintext.
        :param password: str or bytes or bytearray Optional, the password.
        :return: bytes Encrypted data (salt + iv + ciphertext + mac).
        '''
        try:
            data = self._to_bytes(data)
            if self._mode == 'CBC':
                data = pad(data, AES.block_size)

            salt = self._random_bytes(self._salt_len)
            iv = self._random_bytes(self._iv_len)

            aes_key, mac_key = self._keys(salt, password)
            cipher = self._cipher(aes_key, iv)
            ciphertext = cipher.encrypt(data)
            mac = self._sign(iv + ciphertext, mac_key)

            encrypted = salt + iv + ciphertext + mac
            if self.base64:
                encrypted = b64encode(encrypted)
            return encrypted
        except (TypeError, ValueError) as e:
            self._error_handler(e)

    def decrypt(self, data, password = None):
        '''
        Decrypts data using the supplied password or a master key.
        The password is not required if a master key has been set -
        either with `random_key_gen` or with `set_master_key`.
        If a password is supplied, it will be used to create a key with PBKDF2.
        :param data: str or bytes or bytearray The ciphertext.
        :param password: str or bytes or bytearray Optional, the password.
        :return: bytes Plaintext.
        '''
        try:
            data = self._to_bytes(data)
            data = b64decode(data) if self.base64 else data

            salt = data[:self._salt_len]
            iv = data[self._salt_len: self._salt_len + self._iv_len]
            ciphertext = data[self._salt_len + self._iv_len: -self._mac_len]
            mac = data[-self._mac_len:]

            aes_key, mac_key = self._keys(salt, password)
            self._verify(iv + ciphertext, mac, mac_key)

            cipher = self._cipher(aes_key, iv)
            plaintext = cipher.decrypt(ciphertext)
            if self._mode == 'CBC':
                plaintext = unpad(plaintext, AES.block_size)
            return plaintext
        except (TypeError, ValueError, Base64Error) as e:
            self._error_handler(e)

    def encrypt_file(self, path, password = None):
        '''
        Encrypts files using the supplied password or a master key.
        The password is not required if a master key has been set -
        either with `random_key_gen` or with `set_master_key`.
        If a password is supplied, it will be used to create a key with PBKDF2.
        The original file is not modified; a new encrypted file is created.
        :param path: str The file path.
        :param password: str or bytes Optional, the password.
        :return: str The new file path.
        '''
        try:
            salt = self._random_bytes(self._salt_len)
            iv = self._random_bytes(self._iv_len)

            aes_key, mac_key = self._keys(salt, password)
            cipher = self._cipher(aes_key, iv)
            hmac = HMAC.new(mac_key, digestmod = SHA256)
            new_path = path + '.enc'

            with open(new_path, 'wb') as f:
                f.write(salt + iv)
                hmac.update(iv)

                for chunk, is_last in self._file_chunks(path):
                    if self._mode == 'CBC' and is_last:
                        chunk = pad(chunk, AES.block_size)
                    data = cipher.encrypt(chunk)
                    f.write(data)
                    hmac.update(data)

                f.write(hmac.digest())
            return new_path
        except (TypeError, ValueError, IOError) as e:
            self._error_handler(e)

    def decrypt_file(self, path, password = None):
        '''
        Decrypts files using the supplied password or a master key.
        The password is not required if a master key has been set -
        either with `random_key_gen` or with `set_master_key`.
        If a password is supplied, it will be used to create a key with PBKDF2.
        The original file is not modified; a new decrypted file is created.
        :param path: str The file path.
        :param password: str or bytes Optional, the password.
        :return: str The new file path.
        '''
        try:
            with open(path, 'rb') as f:
                salt = f.read(self._salt_len)
                iv = f.read(self._iv_len)
                f.seek(file_size(path) - self._mac_len)
                mac = f.read(self._mac_len)

            aes_key, mac_key = self._keys(salt, password)
            self._verify_file(path, mac, mac_key)
            cipher = self._cipher(aes_key, iv)
            new_path = re_sub(r'\.enc$', '.dec', path)

            with open(new_path, 'wb') as f:
                chunks = self._file_chunks(
                    path,
                    self._salt_len + self._iv_len,
                    self._mac_len
                )
                for chunk, is_last in chunks:
                    data = cipher.decrypt(chunk)

                    if self._mode == 'CBC' and is_last:
                        data = unpad(data, AES.block_size)
                    f.write(data)
            return new_path
        except (TypeError, ValueError, IOError) as e:
            self._error_handler(e)

    def set_master_key(self, key, raw = False):
        '''
        Sets a new master key.
        This key will be used to create the encryption and authentication keys.
        :param key: str or bytes or bytearray The new master key.
        :param encoded: bool Optional, expexts raw bytes; not base64-encoded.
        '''
        try:
            if not raw:
                key = b64decode(key)
            self._master_key = self._to_bytes(key)
        except (TypeError, Base64Error) as e:
            self._error_handler(e)

    def get_master_key(self, raw = False):
        '''
        Returns the master key (or `None` if the key is not set).
        :param raw: bool Optional, returns raw bytes; not base64-encoded.
        :return: bytes The master key.
        '''
        if self._master_key is None:
            self._error_handler(ValueError('The key is not set!'))
        elif not raw:
            return b64encode(self._master_key)
        else:
            return self._master_key

    def random_key_gen(self, key_len = 32, raw = False):
        '''
        Generates a new random key.
        This key will be used to create the encryption and authentication keys.
        :param raw: bool Optional, returns raw bytes; not base64-encoded.
        :return: bytes The new master key.
        '''
        self._master_key = self._random_bytes(key_len)
        if not raw:
            return b64encode(self._master_key)
        return self._master_key

    def _keys(self, salt, password):
        '''
        Derives encryption and authentication keys from a key or password.
        If the password is not null, it will be used to create the keys.
        :raises ValueError: if neither the key or password is set.
        '''
        if password is not None:
            dkey = PBKDF2(
                password,
                salt,
                self._key_len + self._mac_key_len,
                self.key_iterations,
                hmac_hash_module = SHA512
            )
        elif self._master_key is not None:
            dkey = HKDF(
                self._master_key,
                self._key_len + self._mac_key_len,
                salt, SHA256
            )
        else:
            raise ValueError('No password or key specified!')
        return (dkey[:self._key_len], dkey[self._key_len:])

    def _random_bytes(self, size):
        '''
        Creates random bytes; used for IV, salt and key generation.
        '''
        return get_random_bytes(size)

    def _cipher(self, key, iv):
        '''
        Creates an AES object; used for encryption / decryption.
        '''
        return AES.new(key, self._modes[self._mode], IV = iv)

    def _sign(self, ciphertext, key):
        '''
        Computes the MAC of ciphertext; used for authentication.
        '''
        hmac = HMAC.new(key, ciphertext, digestmod = SHA256)
        return hmac.digest()

    def _sign_file(self, path, key):
        '''
        Computes the MAC of ciphertext; used for authentication.
        '''
        hmac = HMAC.new(key, digestmod = SHA256)
        for data, _ in self._file_chunks(path, self._salt_len):
            hmac.update(data)
        return hmac.digest()

    def _verify(self, data, mac, key):
        '''
        Verifies the authenticity of ciphertext.
        :raises ValueError: if the MAC is invalid.
        '''
        hmac = HMAC.new(key, data, digestmod = SHA256)
        hmac.verify(mac)

    def _verify_file(self, path, mac, key):
        '''
        Verifies the authenticity of ciphertext.
        :raises ValueError: if the MAC is invalid.
        '''
        hmac = HMAC.new(key, digestmod = SHA256)
        beg, end = self._salt_len, self._mac_len

        for chunk, _ in self._file_chunks(path, beg, end):
            hmac.update(chunk)
        hmac.verify(mac)

    def _error_handler(self, exception):
        '''
        Handles exceptions (prints the exception by default).
        '''
        print(exception)

    def _file_chunks(self, path, beg = 0, end = 0):
        '''
        A generator that reads a file and yields chunks of data.
        The chunk size should be a multiple of 16 in CBC mode.
        '''
        size = 1024
        end = file_size(path) - end

        with open(path, 'rb') as f:
            pos = (len(f.read(beg)))
            while pos < end:
                size = size if end - pos > size else end - pos
                data = f.read(size)
                pos += len(data)

                yield (data, pos == end)

    def _to_bytes(self, data, encoding = 'utf-8'):
        '''
        Converts unicode strings and byte arrays to byte strings.
        '''
        if hasattr(data, 'encode'):
            data = bytes(data, encoding)
        if type(data) is bytearray:
            data = bytes(data)
        return data

#
#k = KriptaAES()
#
#print("\nWELCOME TO KRIPTA-AES-PYTHON")
#message = """
#
#----------------------------------------------------------------------
#[+] s2c v0.0.1 | Client : 336d5ebc5436534e61d16e63ddfca327
#----------------------------------------------------------------------
#?clLxwpE&&&&HHHHH&&H&&&EdEpddddAApAppZGGUUUXXXXaaaaaaahShhhhhh444V3VVV
#?cL7thpE&&&H0HH&HHH&&EEdAEddApdAAAppZpZUUUUUXXXaaaaaaaSShhhhhhhh4V4V4h
#cl7YCU&HH0$0000H0H&&&&EAEEAdddddEdddAAApppGGUUUXXXaaaaaaaaaShShhhhhhhh
#L7YtwA&$0$0$$D$00H0&&&&E&&EE&E&EEE&EdddAAppppGUUUUUUXXXaaaaaaSaSShhhhh
#LYtChHDD$DDD$DD$$$$$HHHH&&&&&H&&HH0H&EEEEddppppppUUUUXaaXaaaaaaaaaaaaa
#jttoG$8D88888888DDDDD0$0HHH&H&&HH0HH&&&&&EEddAAppGGUGUUUUUUXXXXXaaaaXa
#t%{308888888888888DDDDD$$$$000$0$0$$H0HHH&&&&EdEdppppGUUGUUUUUUUUUUXUU
#tf{a8#####Q#888888888888DD$DDDDDDDDDD$$$HS*:+.+?AdApppppZZGZUUGUGUUGUG
#tCypQgggggQQ#g#QQ88QQ888888888888888DD$L.`      .)?pApZpppppZZZZZZZUZp
#tfIHggggRggRg#ggggg####88QQ8888888888D:::_;:.      ^UAAppppAdAAApppppp
#tfh8gggRRRRRRgggRggggg#QQ#8Q#Q8Q8##8h+++___;:,.    .G&EEEdAddddddpdppA
#%{ZggRRRRRRRRRggRggggg######QQ##Q#V8=__=;+_:::...   ?&&EE&EEEEEEAddddd
#tw$RRRRRRRRRRRRRRRggggg#ggg##gg#8+=***^:,.._^.....  .p&&&&E&&&&&&&&EEd
#%hQRRRRRRRRRRRRRRRRRRggggg#gg#g#&_*)?*:...!^::..... ^&H&HH0HH&&&H&&&&E
#fGgRRRRRRRRRRRRRRRgRggggggggQgggQ)*)<___:.. `.`...._$$$000H&&E&&E&&E&E
#y&RRRRRRRRRRRRRRRgggRgRgggg###g#p**)=^)_+:``...`~:lDD$$$0&EEdppdddddAd
#38RRRRRRRRRRRRgRRRRggggggggg#Qg8<**_;:^;_.`. .`._$8DDD0&EpGUXXUUUUUGGZ
#agRRRRRRRRRRRRRRRRgggggggg##g88x*=**:;:. . ....+F88DD$EpUaShh44hhSaSaa
#pgRRRRRRRRRRRRRRRRRgggg######Q$)**++=:.` ```..:h888D$EpXhVIyyoyoyywI3h
#$d0$&$EARRZRRyRRRgRg{Qg#gggQg#?\==\:~:   ``.~_D888DD&ZXhIy{f%nCCCCffoo
#gRNBBNU0NNBNgyNQ8888n##Q#ggQQc)<=::,.    .*$88888DDHpUhIy{%%%%tCfCCfff
#8gRRRghRRgggU0Wg888$&QQ888$%?)**+^.`    `a$8Q888DH&pa4Iy{%%t%%nn%nnnnf
#ggRRRRRRRRRRRWRRRRRgg88X?))?))*=_,.``   ..^G8D$&dZXa3wy{ntttttn%%%nnCC
#gRRRRRRRRRRRRRRRRRg83)*)))))?+))+....` `...~:+XUah3wy{fnttxt%tttt%tt%%
#RRRRRRRRRRRRRRR$c<*****))))*_=*)+:^:~....,~:~::;){fntxxxxxttt%x  7 FPS
#----------------------------------------------------------------------
#23f498becb757ea1645685cef580e94282587f80c5718209cbbe8aa7332050d3 |1774
#----------------------------------------------------------------------
#
#
#"""
#secret_key = "secret-code-password"
#print("-----------------------------------------------------")
#print("Message : ", message)
#print("-----------------------------------------------------")
#print("secret_key : ", secret_key)
#print("-----------------------------------------------------")
#encrypted_msg = k.encrypt(message, secret_key)
#print("Encrypted-Message : ", encrypted_msg.decode('utf-8'))
#print("-----------------------------------------------------")
#print("Decrypted-Message : ", k.decrypt(encrypted_msg, secret_key).decode('utf-8'))
#print("-----------------------------------------------------")
#
#
#print("len message : ", len(message))
#print("len secret-key : ", len(secret_key))
#print("len encrypted_msg : ", len(encrypted_msg.decode('utf-8')))
#print("len bytes encrypted_msg : ", len(encrypted_msg))
#print("elapsed time : ", time.time() - start)
#



from hashlib import md5
from uuid import uuid4, uuid1
from random import randint



def generate_key(prs: object):
    """
    This metho will just generate a secret key for
    AES encryption/decryption

    We check if the provided key is None before generate
    a new one

    """
    if prs.key is None:
        key = str(randint(0, 99999)) + str(uuid4())
        return "s2c_" + md5(key.encode()).hexdigest()[:4]
    else:
        return prs.key


def parse(prs: object):
    """
    This method will generate a new session

    """
    return {
        "session_id": str(uuid1()) if prs.session_id is None else prs.session_id,
        "session_key":  generate_key(prs),
        "client_id": str(uuid4()) if prs.client_id is None else prs.client_id,
        "ip": prs.ip,
        "port": int(prs.port)
    }




# main.py
# The main script for s2c
# parse input parameters and
import argparse
from s2c.client import Client
from s2c.utils.session import parse
from s2c.settings import HOST, PORT



if __name__ == "__main__":
    # Initialize the arguments
    prs = argparse.ArgumentParser()
    prs.add_argument('-s', '--session_id',
            help='The sesion_id, if noting is provide,it will generate for you',
            type=str, default=None)
    prs.add_argument('-c', '--client_id',
            help='Your id or name in the session, if noting is provide,it will generate for you',
            type=str, default=None)
    prs.add_argument('-k', '--key',
            help='To provide the custom key for the AES encryption',
            type=str, default=None)
    prs.add_argument('-i', '--ip',
            help='The host of the server where websockets will transits',
            type=str, default=HOST)
    prs.add_argument('-p', '--port',
            help='The port of the host',
            type=int, default=PORT)

    # We start our client here
    client = Client(parse(prs.parse_args()))




import json, socket, threading, pyaudio
import time
import base64
from os import (
        path as os_path,
        name as os_name,
        system
)
from s2c.settings import *
from cv2 import (
        resize,
        flip,
        cvtColor,
        COLOR_BGR2GRAY,
        VideoCapture
)
from s2c.utils.camera import ascii_it
from s2c.utils.helpers import get_trace



class Client:
    def __init__(self, session):
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.ip = session["ip"]
        self.port = session["port"]

        self.session_key = session["session_key"]
        self.session_id = session["session_id"]
        self.client_id = session["client_id"]

        self.size = [50, 15]
        self.faces = {}

        while True:
            try:
                self.s.connect((self.ip, self.port))
                break
            except:
                print("[x] Couldn't connect to server {}:{}".format(str(self.ip), str(self.port)))
                time.sleep(1)

        # The Cature for cam
        self.cam = VideoCapture(0)
        self.cam.set(2, self.size[1])
        self.cam.set(3, self.size[0])

        chunk_size = 512
        audio_format = pyaudio.paInt16
        channels = 1
        rate = 10000

        # initialise microphone recording
        self.p = pyaudio.PyAudio()
        self.playing_stream = self.p.open(
                format=audio_format,
                channels=channels,
                rate=rate,
                output=True,
                frames_per_buffer=chunk_size
        )
        self.recording_stream = self.p.open(
                format=audio_format,
                channels=channels,
                rate=rate,
                input=True,
                frames_per_buffer=chunk_size
        )

        self.start_logs()

        # start threads
        receive_thread = threading.Thread(target=self.receive_server_data).start()
        threading.Thread(target=self.send_audio_to_server).start()
        threading.Thread(target=self.print_faces).start()
        self.send_frame_to_server()

    def start_logs(self):
        """
        Just the starting logs

        """
        system('cls' if os_name == 'nt' else 'clear')
        print("[-] s2c started...")
        print(f"[>] session_id : {self.session_id}")
        print(f"[<] client_id : {self.client_id}")
        print(f"[:] session_key : {self.session_key}")
        print("[-] Connected to Server")

    def print_faces(self):
        """
        Just to print faces from the self.faces

        """
        while True:
            system('cls' if os_name == 'nt' else 'clear')
            print("-" * 30)
            print(f"[+] s2c v{version} | session_id : {self.session_id}")
            print("-" * 30)

            to_print = ""
            to_print2 = ""
            for i in range(15):
                try:
                    for index, f in enumerate(self.faces):
                        if index < 3:
                            if f not in to_print:
                                to_print += "client_id: " + f + "\n"
                            to_print += self.faces[f].split("\n")[i] + " | "
                        else:
                            if f not in to_print:
                                to_print += "client_id: " + f + "\n"
                            to_print2 += self.faces[f].split("\n")[i] + " | "

                    to_print += "\n"
                    if len(self.faces) > 3:
                        to_print2 += "\n"

                except Exception as es:
                    pass

            print(to_print)
            print("\n" + "-"*30)
            print(to_print2)

    def receive_server_data(self):
        """
        This method is responsible on receiving the video and the audio
        """
        # print("[+] receiver's Thread started...")

        while True:
            received_msg = self.s.recv(2048)
            try:
                received_msg = json.loads(received_msg.decode("utf-8"))

                if "a" in received_msg:
                    audio_chunk = base64.b64decode(received_msg["a"])
                    self.playing_stream.write(audio_chunk)

                if "v" in received_msg:
                    self.faces[received_msg["i"]] = received_msg["v"]

            except (Exception, json.decoder.JSONDecodeError) as es:
                get_trace()

    def send_audio_to_server(self):
        """
        This method will send audio in a seperate thread

        """
        # print("[+] Audio sender's Thread started...")

        while True:
            try:
                # We get the audio stream (1024 in size)
                audio_data = self.recording_stream.read(512)

                # We send the audio tape
                audio_tape = json.dumps({
                    "i": self.client_id,
                    "s": self.session_id,
                    "a": base64.b64encode(audio_data).decode("utf-8")
                })

                try:
                    self.s.sendall(bytes(audio_tape, encoding="utf-8"))
                except ConnectionResetError as es:
                    get_trace()
            except KeyboardInterrupt as es:
                self.cam.release()
                break

    def send_frame_to_server(self):
        """
        This method will send the frames in a while Loop

        """
        while True:
            try:
                _, img = self.cam.read()
                if _:
                    # We enerate our ascii frame
                    ascii_frame = ascii_it(
                                self.client_id,
                                self.session_id,
                                flip(resize(img, (self.size[0], self.size[1])),1))

                    # We send the frame
                    frame = json.dumps({
                        "i": self.client_id,
                        "s": self.session_id,
                        "v": ascii_frame
                    })
                    # We add our self's fce in the faces object
                    self.faces[self.client_id] = ascii_frame
                    try:
                        # We send through sockets
                        self.s.sendall(bytes(frame, encoding="utf-8"))
                    except ConnectionResetError as es:
                        get_trace()
            except KeyboardInterrupt as es:
                self.cam.release()
                break




from bisect import bisect
from s2c.settings import *

from cv2 import (
        resize,
        flip,
        cvtColor,
        COLOR_BGR2GRAY,
        VideoCapture
)
from os import name as os_name, system
from hashlib import sha256



def generate_frame(client_id, session_id, fps_str, gray_image, characters, indices):
    """
    This method will print the ASCII frame and return the encrypted
    frame using PGP encryption

    Args:
        - fps_str,
        - gray_image,
        - characters,
        - indices,
        - gpg,
        - ks
    """

    string = ''
    for row in gray_image:
        for c in row:
            string += characters[indices[c]]
        string += '\n'
    string = string[:-len(fps_str) - 1] + fps_str

    return string


def get_fps(frames):
    """
    Jut to get the frame per second count

    """
    fps = int(frames // (time.time() - start))

    return '  {} FPS'.format(fps)


def ascii_it(client_id, session_id, image):
    """
    From an image to an ASCII representation with brightnesses

    """

    global frames

    # Convert to grayscale
    gray_image = 255 - cvtColor(image, COLOR_BGR2GRAY)

    # Normalize so that the whole range of characters is used
    upper_limit = gray_image.max() * ((len(characters) + 1) / float(len(characters)))
    lower_limit = gray_image.min()

    bright_div = (global_brightnesses - global_brightnesses.min()) / (
                global_brightnesses.max() - global_brightnesses.min())
    brightnesses = bright_div * (upper_limit - lower_limit) + lower_limit

    frames += 1
    fps_str = get_fps(frames)

    for c in range(gray_image.min(), gray_image.max() + 1):
        indices[c] = bisect(brightnesses, c)

    return generate_frame(client_id, session_id, fps_str, gray_image, characters, indices)


###################################################################################################




# Just the encrypt/decrypt methods
# for kripta_aes
from s2c.modules.security.KriptaAES import KriptaAES


# We instantiate the class only once
kripta_aes = KriptaAES()


def encrypt_aes(key: str, msg: str) -> str:
    """
    This method will just encrypt a message from kripta_aes
    module

    """

    return kripta_aes.encrypt(msg, key).decode("utf-8")

def decrypt_aes(key: str, enc_msg: str) -> str:
    """
    This method will just decrypt an encrypted message from
    kripta_aes module

    """

    return kripta_aes.decrypt(enc_msg, key).decode("utf-8")




# note - gpg needs to be installed first:
# brew install gpg
# apt install gpg

# you may need to also:
# export GPG_TTY=$(tty)
from os import path as os_path, makedirs



def generate_pub_priv_keys(gpg: object, \
                            file_name: str, \
                            passphrase: str, \
                            share_file_name: str="./data/share"):
    """
    This method will use PGP through GnuPG to generate
    a public and private keys that will be use for encrypt
    and decrypt.

    Args:
        - gpg: The GnuPG instance to conserve the same context
        - file_name: the file name where keys will be stored
        - passphrase: The passphrase for encrypt/decrypt
    """
    # if data directory doesn't exist, let's create it
    if not os_path.exists("./data"):
        makedirs("data")

    # Generate key
    key = gpg.gen_key(gpg.gen_key_input(
        passphrase=passphrase,
        key_type="RSA",
        key_length=1024
    ))

    # Create ascii-readable versions of pub / private keys
    # Export it in a single file
    with open(file_name, 'w') as f:
        f.write(gpg.export_keys(key.fingerprint))
        f.write(gpg.export_keys(
            keyids=key.fingerprint,
            secret=True,
            passphrase=passphrase,
        ))

    # We save also the file that should be share for encryption from the peer
    with open(share_file_name, "w") as f:
        f.write(gpg.export_keys(key.fingerprint))


def import_keys(gpg: object, file_name: str) -> list:
    """
    To import presaved keys before encryption/decryption

    Args:
        - gpg: The GnuPG instance to conserve the same context
        - file_name: The path for the presaved keys
    """

    # import keys
    with open(file_name) as f:
        key_data = f.read()

    return [k["fingerprint"] for k in gpg.import_keys(key_data).results]



def encrypt(gpg: object, input_str: str, recipients: list) -> str:
    """
    This method will encrypt input ascii str

    Args:
        - gpg: The GnuPG instance to conserve the same context
        - input_str: The input ASCII text
        - recipients: The list of recipients
    """

    if len(input_str) > 2:
        # encrypt file
        enc = gpg.encrypt(input_str, recipients, always_trust=True)

        # We print an error if there is a problem
        if not enc.ok:
            print(enc.status)
            return ""

        return str(enc)

    return ""


def decrypt(gpg: object, input_str: str, passphrase: str) -> str:
    """
    This method will decrypt input encrypted ascii str

    Args:
        - gpg: The GnuPG instance to conserve the same context
        - input_str: The encrypted input ASCII text
        - passphrase: THe passphrase to decrypt
    """
    if len(input_str) > 2:
        # decrypt file
        dec = gpg.decrypt(input_str, passphrase=passphrase)

        # We print an error if there is a problem
        if not dec.ok:
            print(dec.status)
            return ""

        return str(dec)

    return ""


# How to test that locallyl
#if __name__ == "__main__":
#    import gnupg
#    import time
#    from os import path
#
#    # Create the gpg instance
#    gpg = gnupg.GPG()
#    key_path = "./data/k"
#    passphrase = "darker"
#
#    # We check if the key is present
#    # if not we generate a key pair
#    if not path.exists(key_path):
#        generate_pub_priv_keys(gpg, key_path, passphrase)
#
#    # We import keys pair from the file saved locally
#    ks = import_keys(gpg, key_path)
#
#    input_str = "Example of text !"
#    print("input_str : ", input_str)
#
#    enc = encrypt(gpg, input_str, recipients=ks)
#    print("enc : ", enc)
#
#    dec = decrypt(gpg, enc, passphrase)
#    print("dec : ", dec)
#
#
